package be.iminds.ilabt.jfed.experimenter_gui.tasks;

import be.iminds.ilabt.jfed.experimenter_gui.dialogs.DialogsFactory;
import be.iminds.ilabt.jfed.experimenter_gui.dialogs.SelectAuthoritiesDialog;
import be.iminds.ilabt.jfed.experimenter_gui.editor.EditableRspec;
import be.iminds.ilabt.jfed.experimenter_gui.editor.ExperimentEditorTabFactory;
import be.iminds.ilabt.jfed.experimenter_gui.events.OpenESpecEvent;
import be.iminds.ilabt.jfed.experimenter_gui.tutorials.TutorialViewsFactory;
import be.iminds.ilabt.jfed.experimenter_gui.ui.ExperimentEditorTabHost;
import be.iminds.ilabt.jfed.fedmon.webapi.service.json.Server;
import be.iminds.ilabt.jfed.highlevel.controller.TaskThread;
import be.iminds.ilabt.jfed.highlevel.model.SfaModel;
import be.iminds.ilabt.jfed.highlevel.model.Slice;
import be.iminds.ilabt.jfed.highlevel.tasks.HighLevelTaskFactory;
import be.iminds.ilabt.jfed.highlevel.tasks.RecoverSliceTask;
import be.iminds.ilabt.jfed.highlevel.tasks.RecoverSliceTaskInteraction;
import be.iminds.ilabt.jfed.lowlevel.user.GeniUserProvider;
import be.iminds.ilabt.jfed.lowlevel.testbed_info.TestbedInfoSource;
import be.iminds.ilabt.jfed.lowlevel.api_wrapper.AggregateManagerWrapper;
import be.iminds.ilabt.jfed.lowlevel.authority.finder.AuthorityFinder;
import be.iminds.ilabt.jfed.preferences.JFedPreferences;
import be.iminds.ilabt.jfed.rspec.generator.RSpecGeneratorFactory;
import be.iminds.ilabt.jfed.util.common.GeniUrn;
import com.google.common.eventbus.EventBus;
import javafx.concurrent.Task;
import javafx.stage.Window;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import javax.inject.Inject;
import javax.inject.Provider;
import javax.inject.Singleton;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;

/**
 * User: twalcari
 * Date: 12/1/2014
 * Time: 17:40
 */
@Singleton
public class TasksFactoryImpl implements TasksFactory {
    private final SfaModel sfaModel;
    private final HighLevelTaskFactory hltf;
    private final TaskThread tt;
    private final EventBus eventBus;
    private final Provider<SelectAuthoritiesDialog> selectAuthoritiesDialogProvider;
    private final Provider<TutorialViewsFactory> tutorialViewsFactoryProvider;
    private final ExperimentEditorTabFactory experimentEditorTabFactory;
    private final ExperimentEditorTabHost experimentEditorTabHost;
    private final TestbedInfoSource testbedInfoSource;
    private final AuthorityFinder authorityFinder;
    private final JFedPreferences jFedPreferences;
    private final DialogsFactory dialogsFactory;
    private final GeniUserProvider geniUserProvider;
    private final RSpecGeneratorFactory rSpecGeneratorFactory;

    @Inject
    TasksFactoryImpl(SfaModel sfaModel, HighLevelTaskFactory hltf,
                     TaskThread tt, EventBus eventBus, Provider<TutorialViewsFactory> tutorialViewsFactoryProvider,
                     TestbedInfoSource testbedInfoSource,
                     AuthorityFinder authorityFinder,
                     JFedPreferences jFedPreferences,
                     Provider<SelectAuthoritiesDialog> selectAuthoritiesDialogProvider,
                     ExperimentEditorTabFactory experimentEditorTabFactory,
                     ExperimentEditorTabHost experimentEditorTabHost,
                     GeniUserProvider geniUserProvider,
                     DialogsFactory dialogsFactory,
                     RSpecGeneratorFactory rSpecGeneratorFactory) {
        this.sfaModel = sfaModel;
        this.hltf = hltf;
        this.tt = tt;
        this.eventBus = eventBus;
        this.tutorialViewsFactoryProvider = tutorialViewsFactoryProvider;
        this.testbedInfoSource = testbedInfoSource;
        this.authorityFinder = authorityFinder;
        this.jFedPreferences = jFedPreferences;
        this.selectAuthoritiesDialogProvider = selectAuthoritiesDialogProvider;
        this.experimentEditorTabFactory = experimentEditorTabFactory;
        this.experimentEditorTabHost = experimentEditorTabHost;
        this.dialogsFactory = dialogsFactory;
        this.geniUserProvider = geniUserProvider;
        this.rSpecGeneratorFactory = rSpecGeneratorFactory;
    }

    @Nonnull
    @Override
    public RecoverSliceTask createRecoverSliceTask(@Nonnull GeniUrn sliceUrn) {
        RecoverSliceTaskInteraction guiInteraction = new RecoverSliceTaskGuiInteraction(selectAuthoritiesDialogProvider, eventBus, null);
        return hltf.createRecoverSliceTask(sliceUrn, guiInteraction);
    }

    @Nonnull
    @Override
    public RecoverSliceTask createRecoverSliceTask(@Nonnull Slice slice) {
        RecoverSliceTaskInteraction guiInteraction = new RecoverSliceTaskGuiInteraction(selectAuthoritiesDialogProvider, eventBus, null);
        return hltf.createRecoverSliceTask(slice, guiInteraction);
    }

    @Nonnull
    @Override
    public DeleteDiskImageTask createDeleteDiskImageTask(@Nonnull Server auth, @Nonnull AggregateManagerWrapper.ImageInfo imageInfo) {
        return new DeleteDiskImageTask(auth, imageInfo, hltf);
    }

    @Nonnull
    @Override
    public OpenRequestRspecTask createOpenRequestRspecTask(@Nullable Window parentWindow, @Nonnull File file) throws FileNotFoundException {
        return new OpenRequestRspecTask(parentWindow, file, experimentEditorTabFactory, experimentEditorTabHost);

    }

    @Nonnull
    @Override
    public OpenRequestRspecTask createOpenRequestRspecTask(@Nullable Window parentWindow, @Nonnull InputStream inputStream, long fileLength, @Nonnull String name, @Nullable File defaultSaveFile) {
        return new OpenRequestRspecTask(parentWindow, inputStream, fileLength, name, defaultSaveFile, experimentEditorTabFactory, experimentEditorTabHost);
    }

    @Nonnull
    @Override
    public Task<EditableRspec> createOpenRequestRspecTaskFromString(@Nullable Window parentWindow, @Nonnull String requestRspecString) {
        return new OpenRequestRspecFromStringTask(parentWindow, requestRspecString, experimentEditorTabFactory, experimentEditorTabHost);
    }

    @Nonnull
    @Override
    public OpenRequestRspecTask createOpenRequestRspecTaskFromUrl(@Nullable Window parentWindow, @Nonnull String urlString) throws IOException {
        return new OpenRequestRspecTask(parentWindow, urlString, experimentEditorTabFactory, experimentEditorTabHost);
    }

    @Nonnull
    @Override
    public FetchAllSlicesInfoTask createFetchAllSlicesInfoTask() {
        return new FetchAllSlicesInfoTask(sfaModel, hltf, tt);
    }


    @Nonnull
    @Override
    public OpenTutorialTask createOpenTutorialTask(@Nullable Window dialogOwnerWindow, @Nonnull String urlString) throws IOException {
        return new OpenTutorialTask(dialogOwnerWindow, urlString, tutorialViewsFactoryProvider.get());
    }

    @Nonnull
    @Override
    public OpenESpecTask createOpenESpecTask(@Nullable Window dialogOwnerWindow, @Nonnull OpenESpecEvent event) throws IOException {
        return new OpenESpecTask(event, jFedPreferences, geniUserProvider, dialogsFactory, rSpecGeneratorFactory);
    }
}
